/*
 * Still Grab 0.1
 * 6/21/97
 * 
 * Danny Dulai <nirva@ishiboo.com>
 * 
 */
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <termios.h>
#include <X11/xpm.h>

#ifdef SONY_DSCF1
#  define CAMERANAME "Sony DSC-F1"
#  include "dscf1.h"
#else
#  ifdef CASIO_QV10 
#    define CAMERANAME "Casio QV-10"
#    include "qv10.h"
#  endif
#endif

#include "main.h"
#include "tty.h"
#include "gfx.h"

extern char *optarg;
int getopt();

char		*filename = "grab";
char		*progname;
Display		*display;
int		screen, depth;
Visual		*visual;
Window		window;
XEvent		event;
GC		gc;
XImage		*xim;
Pixmap		background, bgtile, bgtilemask, frame, framemask;
Pixmap		deletexpm, deletexpmmask, retxpm, retxpmmask, quitxpm, quitxpmmask;
Pixmap		slidexpm, slidexpmmask, slide2xpm, slide2xpmmask, hiframe, hiframemask;
u_long		rmask, gmask, bmask, rrotate, grotate, brotate;
u_char		*imagedata;
int		border, bperline, bperpix, rshift, gshift, bshift;
int		curw, curh, totalpics;
Thumbnail	*thumbnails;

enum { START, IDLE, THUMB, RETRIEVE } state = START;
int		onpic=0;

static int savepicnum=0, totalselected=0, lastselected=0;

int
usage()
{
    printf("\n"
	   "StillGrab 1.0.1 -- %s Frame grabber\n"
	   "Usage: %s [-f filename] [-h]\n"
	   "\n"
	   "-h              This help\n"
	   "-f filename     Default filename prefix to save grabs\n"
	   "-T              Path to serial device\n"
	   "-S [h|l]        Speed: h = 38400bps, l = 9600bps\n"
	   "\n", CAMERANAME, progname);
	return 0;
}

void
Exit(int sig)
{
    if(sig)
	printf("Caught signal %d\n",sig);

    CloseCamera();
    
    printf("Quitting...\n");
    exit(0);
}

FILE *
OpenGrabFile()
{
    FILE *fp;
    char name[1024];
    int fd;
    int i = 1000;

    while(i--) {
	strncpy(name, filename, 1000);
	sprintf(name+strlen(name), "%d.jpg", savepicnum++);
	if ((fd = open(name, O_CREAT|O_EXCL|O_WRONLY, 0644)) >= 0) {
	    fp = fdopen(fd, "w");
	    return fp;
	}
    }
    printf("couldn't open grab file\n");
    Exit(0);
    return NULL;
}

void
DoDelete()
{
    int old = totalpics;
    int i;
    for(i=totalpics-1; i >= 0; i--) {
	if(thumbnails[i].selected)
	    RemoveImage(i);
    }
    if(old != totalpics) {
	Tilebg(curw, curh);
	for(i=0; i < totalpics; i++) {
	    DrawFrame(i, thumbnails[i].selected, 0);
	    if(thumbnails[i].thumb)
		DrawThumb(i, 0);
	}
	XCopyArea(display,background,window,gc,0,0,curw,curh,0,0);
    }
}

void HandleButtonPress(XButtonEvent *event)
{
    int x, y, pos;
    x = event->x;
    y = event->y;

    if(y > curh-60) /* down by slide bar */
	return;

    if(y < 130) { /* up by the top buttons */
	if(x > curw/6-87 && x < curw/6+87 && y > 12 && y < 106) { /* RETRIEVE */
	    if(state == IDLE) {
		onpic = 0;
		state = RETRIEVE;
		XCopyArea(display,slidexpm,background,gc,
			  11, 11, 250, 20, curw/2-136+11, curh-51+11);
		XCopyArea(display,slidexpm,window,gc,
			  11, 11, 250, 20, curw/2-136+11, curh-51+11);
	    }
	    return;
	} else if(x > curw*3/6-83+20 && x < curw*3/6+83+20 && y > 12 && y < 111) { /* DELETE */
	    DoDelete();
	    return;
	} else if(x > curw*5/6-65 && x < curw*5/6+65 && y > 3 && y < 126) /* QUIT */
	    Exit(0);
    }
    
    pos = (x / (THUMB_W+50)) + ((y-130) / (THUMB_H+50)) * (curw / (THUMB_W+50));
    if(pos >= totalpics)
	return;

    if(event->state & ControlMask) {

	if(!thumbnails[pos].selected) {
	    thumbnails[pos].selected=1;
	    totalselected++;
	} else {
	    thumbnails[pos].selected=0;
	    totalselected--;
	}
	DrawFrame(pos, thumbnails[pos].selected, 1);
	if(thumbnails[pos].thumb) DrawThumb(pos, 1);

    } else if(event->state & ShiftMask) {

	for(x=0; x < totalpics; x++) /* erase all other highlights */
	    if(x != pos && thumbnails[x].selected) {
		thumbnails[x].selected=0;
		DrawFrame(x, 0, 1);
		if(thumbnails[x].thumb) DrawThumb(x, 1);
	    }
	totalselected=0;

	if(pos > lastselected) {
	    for (x=lastselected; x <= pos; x++) {
		totalselected++;
		if(!thumbnails[x].selected) {
		    thumbnails[x].selected=1;
		    DrawFrame(x, 1, 1);
		    if(thumbnails[x].thumb) DrawThumb(x, 1);
		}
	    }
	} else {
	    for (x=pos; x <= lastselected; x++) {
		totalselected++;
		if(!thumbnails[x].selected) {
		    thumbnails[x].selected=1;
		    DrawFrame(x, 1, 1);
		    if(thumbnails[x].thumb) DrawThumb(x, 1);
		}
	    }
	}
	lastselected = pos;

    } else if(totalselected == 1 && lastselected == pos) {

	lastselected = 0;
	totalselected=0;
	thumbnails[pos].selected=0;
	DrawFrame(pos, 0, 1);
	if(thumbnails[pos].thumb) DrawThumb(pos, 1);
	return;

    } else {
	lastselected = pos;
	for(x=0; x < totalpics; x++) /* erase all other highlights */
	    if(x != pos && thumbnails[x].selected) {
		thumbnails[x].selected=0;
		DrawFrame(x, 0, 1);
		if(thumbnails[x].thumb) DrawThumb(x, 1);
	    }
	if(!thumbnails[pos].selected) { /* if this is new highlight, highlight it */
	    thumbnails[pos].selected=1;
	    DrawFrame(pos, 1, 1);
	    if(thumbnails[pos].thumb) DrawThumb(pos, 1);
	}
	totalselected=1;
    }

}

int
highbit(unsigned long ul)
{
    /* returns position of highest set bit in 'ul' as an integer (0-31),
       or -1 if none */

    int i;
    unsigned long hb;

    hb = 0x80;
    hb = hb << 24;   /* hb = 0x80000000UL */
    for (i=31; ((ul & hb) == 0) && i>=0;  i--, ul<<=1);
    return i;
}

void
main(int argc, char *argv[])
{
    struct timeval wait;
    KeySym keysym;
    XComposeStatus compose;
    char keybuffer[11], *devpath = RSPORT;
    int ch, i, speed = DEFAULTBAUD;

    progname = argv[0];

    signal(SIGBUS, Exit);
    signal(SIGINT, Exit);
    signal(SIGHUP, Exit);
    signal(SIGQUIT, Exit);
    signal(SIGSEGV, Exit);
    signal(SIGTERM, Exit);

    while ((ch = getopt(argc, argv, "S:T:f:h")) != EOF)
	switch(ch) {
	    case 'S':
		switch(optarg[0]){
		    case 'h':
			speed = B38400;
			break;
		    case 'l':
			speed = B9600;
			break;
		}
		break;
	    case 'T':
		devpath = optarg;
		break;
	    case 'f':
		filename = optarg;
		break;
	    case '?':
	    case 'h':
	    default:
		usage();
		exit(0);
	}

    if(InitCamera(devpath, speed) < 0) {
      	printf("Camera not responding\n");
	exit(1);
    }

    totalpics = GetPicCount();
    printf("pictures = %d\n", totalpics);
    InitThumbnails(totalpics);

    if((display = XOpenDisplay(NULL)) == NULL) {
	printf("Can't open X display\n");
	exit(1);
    }
    screen = DefaultScreen(display);
    visual = DefaultVisual(display,screen);
    depth = DisplayPlanes(display,screen);

    rmask = visual->red_mask;
    gmask = visual->green_mask;
    bmask = visual->blue_mask;
    rshift = 7 - highbit(rmask);
    gshift = 7 - highbit(gmask);
    bshift = 7 - highbit(bmask);

    for(i=0; i < totalpics; i++) {
	xim = XCreateImage (display, visual, depth, ZPixmap, 0, NULL, THUMB_W, THUMB_H, 32, 0);
	if(!xim) {
	    fprintf(stderr,"Fatal: Could not Create XImage\n");
	    exit(1);
	}
	bperline = xim->bytes_per_line;
	bperpix  = xim->bits_per_pixel;
	border   = xim->byte_order;
	if(bperpix != 32) {
	    fprintf(stderr, "Fatal: No support to handle %d-bit TrueColor/DirectColor displays (only 32bpp now)\n", bperpix);
	    exit(1);
	}
	xim->data = malloc(THUMB_H * bperline);
	thumbnails[i].thumbxim = xim;
    }

    InitWindow(argc, argv);

    XFlush(display);
    XMapWindow(display,window);

    while(1) {
	if(XPending(display)) {
	    XNextEvent(display, &event);
	    switch(event.type) {
		case ConfigureNotify:
		    {
			Window root;
			int x, y, i;
			unsigned int c, r, bdr_width, depth;

			XGetGeometry(display,window,&root,&x,&y,&c,&r,&bdr_width,&depth);
			if(c != curw || r != curh) {
			    curw = c; curh = r;

			    if(background)
				XFreePixmap(display, background);
			    background = XCreatePixmap(display, window, curw, curh, XDefaultDepth(display,screen));

			    Tilebg(curw, curh);
			    for(i = 0; i < totalpics; i++) {
				DrawFrame(i, thumbnails[i].selected, 0);
				if(thumbnails[i].thumb)
				    DrawThumb(i, 0);
			    }
			    if(state == START)
				state = THUMB;
			}
		    }
		    break;
		case Expose:
		    HandleExposeEvent((XExposeEvent *) &event);
		    break;
		case ButtonPress:
		    HandleButtonPress((XButtonEvent *) &event);
		    break;
		case KeyPress:
		    XLookupString(&event.xkey, keybuffer, 10, &keysym, &compose);
		    switch(keysym) {
			case XK_a:
			    if(((XKeyEvent*)&event)->state & ControlMask) {
				for(i=0; i < totalpics; i++)
				    if(!thumbnails[i].selected) {
					DrawFrame(i, 1, 1);
					if(thumbnails[i].thumb) DrawThumb(i, 1);
					thumbnails[i].selected=1;
				    }
				totalselected=totalpics;
			    }
			    break;
			case XK_d: /* delete */
			    if(((XKeyEvent*)&event)->state & ControlMask) {
				DoDelete();
			    }
			    break;
			case XK_Return:
			    if(state == IDLE) {
				onpic = 0;
				state = RETRIEVE;
				XCopyArea(display,slidexpm,background,gc,
					  11, 11, 250, 20, curw/2-136+11, curh-51+11);
				XCopyArea(display,slidexpm,window,gc,
					  11, 11, 250, 20, curw/2-136+11, curh-51+11);
			    }
			    break;
			default:
			    break;
		    }
		    break;
		default:
		    break;
	    }
	} else {
	    switch(state) {
		case START:
		    break;
		case IDLE:
		    wait.tv_sec=0;
		    wait.tv_usec=10000; 
		    select(0, NULL, NULL, NULL, &wait);
		    break;
		case RETRIEVE:
		    {
			static int size=0;
			int i;
			if(onpic < totalpics) {
			    if(thumbnails[onpic].selected) {
				switch((i=GetPicture())) {
				    case -1:
					printf("F1finfo() failure\n");
					break;
				    case -2:
					printf("F1fopen() failure\n");
					break;
				    case -3:
					printf("F1fread() failure\n");
					break;
				    case 0:
					totalselected--;
					thumbnails[onpic-1].selected=0;
					DrawFrame(onpic-1, 0, 1);
					if(thumbnails[onpic-1].thumb) DrawThumb(onpic-1, 1);
					XCopyArea(display,slidexpm,background,gc,
						  11, 11, 250, 20, curw/2-136+11, curh-51+11);
					XCopyArea(display,slidexpm,window,gc,
						  11, 11, 250, 20, curw/2-136+11, curh-51+11);
					break;
				    default:
					if(i >= 2) {
					    size = i-2;
					} else if(i <= -3) {
					    i = (-(i+3))*250/size;
					    XCopyArea(display,slide2xpm,background,gc,
						      0, 0, i, 20, curw/2-136+11, curh-51+11);
					    XCopyArea(display,slide2xpm,window,gc,
						      0, 0, i, 20, curw/2-136+11, curh-51+11);
					}
					break;
				}
			    } else 
				onpic++;
			} else
			    state = IDLE;
		    }
		    break;
		case THUMB:
		    if(onpic < totalpics) {
			switch(GetThumbnail()) {
			    case -1:
				printf("F1finfo() failure\n");
				break;
			    case -2:
				printf("F1fopen() failure\n");
				break;
			    case -3:
				printf("F1fread() failure\n");
				break;
			    case 0:
				{
				    int w;
				    w = onpic*250/totalpics;
				    XCopyArea(display,slide2xpm,background,gc,
					      0, 0, w, 20, curw/2-136+11, curh-51+11);
				    XCopyArea(display,slide2xpm,window,gc,
					      0, 0, w, 20, curw/2-136+11, curh-51+11);

				    Data2XImage(thumbnails[onpic-1].thumb,
						thumbnails[onpic-1].thumblen,
						thumbnails[onpic-1].thumbxim,
						thumbnails[onpic-1].rotate == 8 ? 1 : 0);
				    DrawThumb(onpic-1, 1);
				}
				break;
			    case 1:
				break;
			}
		    } else
			state = IDLE;
		    break;
	    }
	}
    }
}
